/*
 * Board.cs
 * kenrick@uaa.alaska.edu
 * 
 * This class stores the board using a 2D array and contains
 * logic to determine if a given move is valid, or if
 * there is a win or capture to actually implement a move.
 * 
 */
using System;
using System.Collections.Generic;
using System.Text;

namespace FitchnealServer
{
    class Board
    {
        public const int EMPTY = 0;                 // Constants to store pieces on the board
        public const int DEFENDER = 1;
        public const int ATTACKER = 2;
        public const int KING = 3;
        public const int SIZE = 9;
        private int[,] board = new int[SIZE, SIZE];

        public Board()
        {               
        }

        public Board(Board inBoard)
        {
            for (int i = 0; i < SIZE; i++)
                for (int j = 0; j < SIZE; j++)
                    board[i, j] = inBoard.board[i,j]; 
        }

        // Sets the board up in the initial configuration
        public void initialize()
        {
            for (int i = 0; i < SIZE; i++)
                for (int j = 0; j < SIZE; j++)
                    board[i, j] = EMPTY; 

            board[0, 0] = ATTACKER;
            board[8, 0] = ATTACKER;
            board[8, 8] = ATTACKER;
            board[0, 8] = ATTACKER;
            board[4, 1] = ATTACKER;
            board[1, 4] = ATTACKER;
            board[7, 4] = ATTACKER;
            board[4, 7] = ATTACKER;
            for (int y = 3; y <= 5; y++)
            {
                board[0, y] = ATTACKER;
                board[8, y] = ATTACKER;
                board[y, 0] = ATTACKER;
                board[y, 8] = ATTACKER;
            }
            for (int x = 2; x <= 6; x++)
                board[x, 4] = DEFENDER;
            for (int y = 2; y <= 6; y++)
                board[4, y] = DEFENDER;
            board[4, 4] = KING;            
        }
        
        // return a piece at coordinate x,y  (in array terms, not board index typed in as a move)
        public int getPiece(int x, int y)
        {
            return board[x,y];
        }

        // Determins if a move on the board is valid.
        // Returns true if valid, false if not.
        public bool validMove(int x1, int y1, int x2, int y2, int whoseturn)
        {
            // Coordinate is out of bounds
            if ((x1 < 0) || (x1 >= SIZE) ||
                (x2 < 0) || (x2 >= SIZE) ||
                (y1 < 0) || (y1 >= SIZE) ||
                (y2 < 0) || (y2 >= SIZE))
                return false;
            // Trying to move a piece that is not yours
            if (whoseturn == ATTACKER)
            {
                if (board[x1, y1] != whoseturn) return false;
            }
            else
            {
                if ((board[x1, y1] != Board.KING) && (board[x1, y1] != Board.DEFENDER))
                    return false;
            }
            // Trying to move to a spot that is occupied
            if (board[x2, y2] != Board.EMPTY) return false;                
            // Trying to move to the throne and not the king
            if ((x2 == 4) && (y2 == 4) && (board[x1, y1] != Board.KING)) return false;
            
            // Determine if moving horizontally or vertically
            if (x1 == x2)
            {
                if (y1 == y2) return false;
                // Horizontal move
                if (y1 < y2)
                {
                    for (int y = y1+1; y <= y2; y++)
                    {
                        if (board[x1, y] != Board.EMPTY) return false;
                    }
                }
                else
                {
                    for (int y = y1 - 1; y >= y2; y--)
                    {
                        if (board[x1, y] != Board.EMPTY) return false;
                    }
                }
            }
            else if (y1 == y2)
            {
                if (x1 == x2) return false;
                // Vertical move
                if (x1 < x2)
                {
                    for (int x = x1 + 1; x <= x2; x++)
                    {
                        if (board[x, y1] != Board.EMPTY) return false;
                    }
                }
                else
                {
                    for (int x = x1 - 1; x >= x2; x--)
                    {
                        if (board[x, y1] != Board.EMPTY) return false;
                    }
                }
            }
            else
            {
                return false;   // x1 must be x2 or y1 must be y2
            }

            // Passed all our tests, it's a valid move
            return true;
        }

        // Implements the move.
        // Assumes the move is valid.
        // Returns number of pieces captured.
        // Returns -1 if whoseturn just won the game.
        public int makeMove(int x1, int y1, int x2, int y2, int whoseturn)
        {
            int numcaptures = 0;

            // Put piece at target
            board[x2, y2] = board[x1, y1];

            // Remove piece from source
            board[x1, y1] = Board.EMPTY;

            // Check for removing any opponent pieces
            if (whoseturn == Board.ATTACKER)
            {
                // Check left
                if (x2 >= 2)
                {
                    if ((board[x2 - 1, y2] == Board.DEFENDER) && (board[x2 - 2, y2] == Board.ATTACKER))
                    {
                        numcaptures++;
                        board[x2 - 1, y2] = Board.EMPTY;
                    }
                }
                // Check up
                if (y2 >= 2)
                {
                    if ((board[x2, y2 - 1] == Board.DEFENDER) && (board[x2, y2 - 2] == Board.ATTACKER))
                    {
                        numcaptures++;
                        board[x2, y2 - 1] = Board.EMPTY;
                    }
                }
                // Check right
                if (x2 < Board.SIZE - 2)
                {
                    if ((board[x2 + 1, y2] == Board.DEFENDER) && (board[x2 + 2, y2] == Board.ATTACKER))
                    {
                        numcaptures++;
                        board[x2 + 1, y2] = Board.EMPTY;
                    }
                }
                // Check down
                if (y2 < Board.SIZE - 2)
                {
                    if ((board[x2, y2 + 1] == Board.DEFENDER) && (board[x2, y2 + 2] == Board.ATTACKER))
                    {
                        numcaptures++;
                        board[x2, y2 + 1] = Board.EMPTY;
                    }
                }
            }
            else if (board[x2, y2] == Board.KING)
            {
                // No captures possible with the king so nothing here
            }
            else    // ** DEFENDER MOVING check for captures
            {
                // Check left
                if (x2 >= 2)
                {
                    if ((board[x2 - 1, y2] == Board.ATTACKER) && (board[x2 - 2, y2] == Board.DEFENDER))
                    {
                        numcaptures++;
                        board[x2 - 1, y2] = Board.EMPTY;
                    }
                }
                // Check up
                if (y2 >= 2)
                {
                    if ((board[x2, y2 - 1] == Board.ATTACKER) && (board[x2, y2 - 2] == Board.DEFENDER))
                    {
                        numcaptures++;
                        board[x2, y2 - 1] = Board.EMPTY;
                    }
                }
                // Check right
                if (x2 < Board.SIZE - 2)
                {
                    if ((board[x2 + 1, y2] == Board.ATTACKER) && (board[x2 + 2, y2] == Board.DEFENDER))
                    {
                        numcaptures++;
                        board[x2 + 1, y2] = Board.EMPTY;
                    }
                }
                // Check down
                if (y2 < Board.SIZE - 2)
                {
                    if ((board[x2, y2 + 1] == Board.ATTACKER) && (board[x2, y2 + 2] == Board.DEFENDER))
                    {
                        numcaptures++;
                        board[x2, y2 + 1] = Board.EMPTY;
                    }
                }
            }
            if (wonGame(x2, y2, whoseturn))
                return -1;
            return numcaptures;
        }

        // Returns true if whomoved won the game,
        // false otherwise.
        private bool wonGame(int x, int y, int whomoved)
        {
            if (whomoved == Board.DEFENDER)
            {
                // Defender just moved, check if won
                if (board[x, y] == Board.KING)
                {
                    if ((x == 0) || (x == Board.SIZE - 1) || (y == 0) || (y == Board.SIZE - 1))
                        return true;
                }
            }
            else
            {
                // Attacker just moved, check if won
                // Search for an adjacent king
                if (x > 0)
                    if (board[x - 1, y] == Board.KING)
                        return kingSurrounded(x - 1, y);
                if (y > 0)
                    if (board[x, y - 1] == Board.KING)
                        return kingSurrounded(x, y - 1);
                if (x < Board.SIZE-1)
                    if (board[x + 1, y] == Board.KING)
                        return kingSurrounded(x + 1, y);
                if (y < Board.SIZE - 1)
                    if (board[x, y + 1] == Board.KING)
                        return kingSurrounded(x, y + 1);
            }
            return false;
        }

        // Returns true if the king is surrounded by attackers, and hence loses the game.
        private bool kingSurrounded(int x, int y)
        {
            int numSurrounds = 0;

            // Surrounded on left?  If x==0 short circuits the IF statement so we won't go off the board with board[x-1,y]
            // although in theory that should be impossible, since the defender would have won if the king is at column 0
            if ((x == 0) || (board[x - 1, y] == Board.ATTACKER) || ((x == 5) && (y == 4)))
                numSurrounds++;
            // Surrounded on top?
            if ((y == 0) || (board[x, y - 1] == Board.ATTACKER) || ((x == 4) && (y == 5)))
                numSurrounds++;
            // Surrounded on right?
            if ((x == Board.SIZE-1) || (board[x + 1, y] == Board.ATTACKER) || ((x == 3) && (y == 4)))
                numSurrounds++;
            // Surrounded on bottom?
            if ((y == Board.SIZE - 1) || (board[x, y + 1] == Board.ATTACKER) || ((x == 4) && (y == 3)))
                numSurrounds++;

            if (numSurrounds == 4) return true;
            return false;
        }

        // Returns number of attacker's pieces
        public int numAttackerPieces()
        {
            int numPieces = 0;
            for (int i = 0; i < SIZE; i++)
                for (int j = 0; j < SIZE; j++)
                    if (board[i, j] == Board.ATTACKER) numPieces++;
            return numPieces;
        }

        // Returns number of attacker's pieces
        public int numAttackerCaptured()
        {
            return 20 - numAttackerPieces();
        }

        // Returns number of defender's pieces
        public int numDefenderPieces()
        {
            int numPieces = 0;
            for (int i = 0; i < SIZE; i++)
                for (int j = 0; j < SIZE; j++)
                    if (board[i, j] == Board.DEFENDER) numPieces++;
            return numPieces;
        }

        // Returns number of defender's pieces, not including the king
        public int numDefenderCaptured()
        {
            return 8 - numDefenderPieces();
        }
    }
}
