/*
 * Fitchneal server v 0.2
 * 9/22/06 Kenrick Mock
 * If you find any errors send them to kenrick@uaa.alaska.edu
 * 
 * You are also welcome to use the source code and extend this program
 * any way you see fit. Maybe you'd like to add mouse input for
 * the human player?  If you make a significant contribution please
 * send it to me and I'll post it for all to use.
 * 
 * This program implements a server to play the game of Fitchneal.
 * The rules it uses are:
 * 
 *   Defender wins if king reaches the edge
 *   King captured if surrounded on all sides by attackers,
 *     or surrounded on three sides by attackers and the throne
 *   King cannot be used to capture attacker's pieces
 *   The king is the only piece that can land on the throne
 *   All pieces can move over the throne
 * 
 * If network play is selected then the network player should
 * connect to the IP of this machine on port 8112.
 * The first network player to connect will be the attacker
 * and the second the defender.  The server will send back a
 * String of "attacker" or "defender" to tell the client which
 * player it is.  A simple way to test the network interface is
 * to simply telnet to the server, e.g. from a command line:
 *    telnet 137.229.156.33 8112
 * and then type moves into the telnet window.
 * 
 * Network clients send moves in the form  X1 Y1 X2 Y2
 * followed by a newline or carriage return.   The server will send the opponent's
 * move to the client in the same format, but followed by a carriage return
 * and line feed.
 * 
 * It is the responsibility of the client to ensure that it makes only
 * valid moves and to keep track of the board state itself.  Any invalid
 * move by a network client results in immediate disqualification.
 * 
 * 10/28/06  Added pieces captured and move log
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Net;
using System.Net.Sockets;

namespace FitchnealServer
{
    public partial class Fitchneal : Form
    {
        private Board mainBoard = new Board();
        private int whoseturn = Board.ATTACKER;
        private int movenum = 1;
        private bool gameOver = false;
        private List<String> moveList;
        public Socket mainSocket = null;   // These should be private
        public Socket sockDefender = null;  // These should be private
        public Socket sockAttacker = null;  // These should be private
        private DateTime startTime;  // For measuring time required to move

        public Fitchneal()
        {
            InitializeComponent();
            
            // Listen on port 8112 for incoming connections
            IPEndPoint ipep = new IPEndPoint(IPAddress.Any, 8112);
            mainSocket = new Socket(AddressFamily.InterNetwork,
                            SocketType.Stream, ProtocolType.Tcp);
            mainSocket.Bind(ipep);
            mainSocket.Listen(10);

            init();
        }

        // Starts a new game and re-initializes the board
        // and the sockets
        private void init()
        {
            this.gameOver = false;
            listMoves.Items.Clear();
            lblAttackersCaptured.Text = "0";
            lblDefendersCaptured.Text = "0";
            lblLastMove.Text = "";
            mainBoard.initialize();
            moveList = new List<String>();
            movenum = 1;
            gameOver = false;
            if (sockDefender != null)
            {
                sockDefender.Disconnect(false);
                sockDefender.Close();
                sockDefender = null;
            }
            if (sockAttacker != null)
            {
                sockAttacker.Disconnect(false);
                sockAttacker.Close();
                sockAttacker = null;
            }
        }

        // This subroutine draws the game board
        private void pboxBoard_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;

            for (int x=0; x<9; x++)
                for (int y = 0; y < 9; y++)
                {
                    switch (mainBoard.getPiece(x, y))
                    {
                        case Board.EMPTY :
                            g.DrawImage(pboxBlank.Image, x * 40, y * 40);
                            break;
                        case Board.ATTACKER :
                            g.DrawImage(pboxAttacker.Image, x * 40, y * 40);
                            break;
                        case Board.DEFENDER :
                            g.DrawImage(pboxDefender.Image, x * 40, y * 40);
                            break;
                        case Board.KING :
                            g.DrawImage(pboxKing.Image, x * 40, y * 40);
                            break;
                    }
                }
        }

        private void Fitchneal_Load(object sender, EventArgs e)
        {
            startTime = DateTime.Now;
        }

        // Handle user entering a move in the textbox        
        private void btnMove_Click(object sender, EventArgs e)
        {
            handleMove(txtMove.Text, 0);
        }

        // Move = string containing coordinates of move
        // caller = 0 if human entered the move, 1 if from the network socket
        private void handleMove(string move, int caller)
        {
            char[] delimiterChars = { ' ', '\t' };
            string[] coords = move.Split(delimiterChars);
            int x1, x2, y1, y2;

            if (coords.Length != 4)
            {
                // Let human try again if made a typo
                if (caller == 0)
                {
                    MessageBox.Show("Error, input 4 coordinates.  Move received: " + move);
                    return;
                }
                else
                {
                    // If network player involved, end game immediately
                    toggleTurn();
                    showGameOverMessage("ERROR, four coordinates not received.  Player forfeits!  Bad move was: ", move);
                    return;
                }
            }

            moveList.Add(Convert.ToString(this.movenum) + ". " + move);         // Save move made in the move list
            listMoves.Items.Add(Convert.ToString(this.movenum) + ". " + move);  // Save move in GUI listbox
            lblLastMove.Text = move;    // Show the user what move was made
            sendLastMove(move);         // Send move to any network player

            // Show time taken
            DateTime stopTime = DateTime.Now;
            TimeSpan duration = stopTime - startTime;
            lblLastMove.Text += ".   Time : " + duration.Seconds.ToString() + "s.  ";

            x1 = Convert.ToInt32(coords[0]) - 1;        // Subtract 1 to map board coordinates to array coords
            y1 = Convert.ToInt32(coords[1]) - 1;
            x2 = Convert.ToInt32(coords[2]) - 1;
            y2 = Convert.ToInt32(coords[3]) - 1;
            if (mainBoard.validMove(x1, y1, x2, y2, whoseturn))
            {
                int captures = mainBoard.makeMove(x1, y1, x2, y2, whoseturn);
                this.pboxBoard.Invalidate();
                if (captures > 0)
                {
                    this.lblLastMove.Text += (captures.ToString() + " pieces were captured. ");
                    toggleTurn();
                }
                else if (captures == -1)
                {
                    //toggleTurn();
                    showGameOverMessage("Game over!  Winning Move: ", move);
                    return;
                }
                else
                {
                    this.lblStatus.Text = "";
                    toggleTurn();
                }
                lblAttackersCaptured.Text = Convert.ToString(mainBoard.numAttackerCaptured());
                lblDefendersCaptured.Text = Convert.ToString(mainBoard.numDefenderCaptured());
            }
            else
            {
                if (caller == 0)
                {
                    MessageBox.Show("Error, invalid move entered: " + move);
                    return;
                }
                else
                {
                    // If network player involved, end game immediately
                    toggleTurn();
                    showGameOverMessage("Invalid move entered.  The player forfeits!  The bad move was:", move);
                    return;
                }
            }
            this.txtMove.SelectAll();
            this.txtMove.Focus();
            Application.DoEvents();

            // Enable timer to get next move from appropriate place (Form or Network)
            // so the next move is fetched outside the scope of this
            // method.  A separate thread for the sockets would be better... this is
            // the cheap version in lieu of our own threads
            timerNetwork.Enabled = true;
        }
       
        // If the user presses enter in the textbox, make it like the
        // user clicked on the submit button
        private void txtMove_KeyPress_1(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 13)
            {
                btnMove_Click(sender, e);
                e.Handled = true;
            }
        }

        // Game over, someone won
        private void showGameOverMessage(string message, string move)
        {
            MessageBox.Show(message + " " + move);
            if (whoseturn == Board.ATTACKER)
                this.lblStatus.Text = "The attacker has won!";
            else
                this.lblStatus.Text = "The defender has won!";
            gameOver = true;
            txtMove.Enabled = false;
            btnMove.Enabled = false;
            saveGameLog();
        }

        // Switch who's turn it is
        private void toggleTurn()
        {
            movenum++;
            if (whoseturn == Board.ATTACKER)
            {
                whoseturn = Board.DEFENDER;
                lblStatus.Text += "It is move " + movenum.ToString() + " and the defender's turn.";
            }
            else
            {
                whoseturn = Board.ATTACKER;
                lblStatus.Text += "It is move " + movenum.ToString() + " and the attacker's turn.";
            }
        }

        // Save game log to disk for human analysis if desired
        private void saveGameLog()
        {
            SaveFileDialog saveFileDialog1 = new SaveFileDialog();

            saveFileDialog1.Title = "Save Log File";
            saveFileDialog1.FileName = "log.txt";
            saveFileDialog1.Filter = "txt files (*.txt)|*.txt|All files (*.*)|*.*";
            saveFileDialog1.FilterIndex = 2;
            saveFileDialog1.RestoreDirectory = true;

            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                TextWriter wText = new StreamWriter(saveFileDialog1.FileName);
                wText.WriteLine(lblStatus.Text);
                foreach (string s in moveList)
                {
                    wText.WriteLine(s);
                }
                wText.WriteLine("Attackers captured: " + lblAttackersCaptured.Text);
                wText.WriteLine("Defenders captured: " + lblDefendersCaptured.Text);
                wText.Close();
            }
        }

        // Start a new game with either the human vs. human using the Form
        private void rbHumanVsHuman_CheckedChanged(object sender, EventArgs e)
        {
            if (rbHumanVsHuman.Checked)
            {
                init();
                pboxBoard.Invalidate();
                lblStatus.Text = "It is move 1 and the attacker's turn.";
                enableMoveInput();
                startTime = DateTime.Now;
            }
        }
        
        // New game with network player as attacker, human using the form as defender
        private void rbNetworkVsHuman_CheckedChanged(object sender, EventArgs e)
        {
            if (rbNetworkVsHuman.Checked)
            {
                init();
                GetPlayers frm = new GetPlayers(0, this);
                frm.ShowDialog();
                pboxBoard.Invalidate();
                disableMoveInput();
                timerNetwork.Enabled = true;
            }
        }

        // New game with the human on the form as attacker, network player as defender
        private void rbHumanVsNetwork_CheckedChanged(object sender, EventArgs e)
        {
            if (rbHumanVsNetwork.Checked)
            {
                init();
                GetPlayers frm = new GetPlayers(1, this);
                frm.ShowDialog();
                pboxBoard.Invalidate();
                disableMoveInput();
                timerNetwork.Enabled = true;
            }
        }

        // New game with two network players
        private void rbNetworkVsNetwork_CheckedChanged(object sender, EventArgs e)
        {
            if (rbNetworkVsNetwork.Checked)
            {
                init();
                GetPlayers frm = new GetPlayers(2, this);
                frm.ShowDialog();
                pboxBoard.Invalidate();
                disableMoveInput();
                timerNetwork.Enabled = true;
            }
        }

        // Get a move from the network player.
        // Since the socket receives data in chunks we have to put together all chunks until
        // we get a newline or carriage return, then stop reading and use the data input as a move
        private string getMove(Socket sock)
        {
            if (whoseturn == Board.ATTACKER)
                lblStatus.Text = "It is move " + movenum.ToString() + ".  Waiting for the attacker's move.";
            else
                lblStatus.Text = "It is move " + movenum.ToString() + ".  Waiting for the defender's move.";
            Application.DoEvents();
            byte[] data = new byte[1024];
            // Keep reading until we get a newline
            bool gotNewline = false;
            string move = "";
            while (!gotNewline)
            {
                string chunk = "";
                int recv = sock.Receive(data);
                if (recv > 0)
                {
                    chunk = Encoding.ASCII.GetString(data, 0, recv);
                    if ((chunk.IndexOf('\r') >= 0) || (chunk.IndexOf('\n') >= 0))
                    {
                        gotNewline = true;
                    }
                    move += chunk;
                }
            }
            string filteredMove = "";
            char[] cArray = move.ToCharArray();
            // Only keep spaces and letters for the benefit
            // of humans who may type in backspaces or tabs, etc.
            for (int i = 0; i < cArray.Length; i++)
            {
                char c = cArray[i];
                if (((c >= '0') && (c <= '9')) || (c==' '))
                    filteredMove += c.ToString();                
            }
            txtMove.SelectAll();
            txtMove.Focus();
            return filteredMove;
        }

        // Determine if we need to send the last move to a network player
        private void sendLastMove(string move)
        {
            byte[] data = new byte[1024];
            data = Encoding.ASCII.GetBytes(move + "\r\n");
            if ((whoseturn == Board.ATTACKER) && (sockDefender != null))
            {
                sockDefender.Send(data, data.Length, SocketFlags.None);
            }
            else if ((whoseturn == Board.DEFENDER) && (sockAttacker != null))
            {
                sockAttacker.Send(data, data.Length, SocketFlags.None);
            }
        }

        // If waiting for the network player, disable the textbox and button for the user
        // on the form
        private void disableMoveInput()
        {
            btnMove.Enabled = false;
            lblEnterMove.Enabled = false;
            txtMove.Enabled = false;
        }

        // If it's a humans turn, enable the textbox and button on the form
        private void enableMoveInput()
        {
            btnMove.Enabled = true;
            lblEnterMove.Enabled = true;
            txtMove.Enabled = true;
        }

        // This timer allows the program to check to see whose move
        // it is after the previous player has already made their move.
        // This allows all moves to be handled at the top of the
        // event loop instead of a big string of function calls for
        // each player's turn.  Ideally this should be implemented
        // with multithreading.
        //
        // The timer goes off 100ms after the previous player has
        // had their move handled.  It then either waits for the
        // network player to submit their move, or it allows the
        // user to enter their move.
        private void timerNetwork_Tick(object sender, EventArgs e)
        {
            startTime = DateTime.Now;

            timerNetwork.Enabled = false;
            // Figure out where to get the next move
            if (whoseturn == Board.ATTACKER)
            {
                if (sockAttacker != null)
                {
                    disableMoveInput();
                    string move = getMove(sockAttacker);
                    handleMove(move, 1);
                }
                else
                {
                    enableMoveInput();
                }
            }
            else
            {
                if (sockDefender != null)
                {
                    disableMoveInput();
                    string move = getMove(sockDefender);
                    handleMove(move, 1);
                }
                else
                {
                    enableMoveInput();
                }
            }
        }

        private void btnMove_Click_1(object sender, EventArgs e)
        {
            handleMove(txtMove.Text, 0);
        }
    }    
}